#ifndef	_H_CONVERT
#define	_H_CONVERT

#define	SECRET				98

#define	CTYPE_INVALID		0
#define	CTYPE_DOOM			1
#define	CTYPE_ZDOOM			2
#define	CTYPE_ZDOOMTC		3
#define	CTYPE_BSTC			4
#define	CTYPE_EDGE			5

#define	LINE_LEFT			1
#define	LINE_RIGHT			2
#define	LINE_ABOVE			3
#define	LINE_BELOW			4

#define	DOOR_UNLOCKED		1
#define	DOOR_SILVER			2
#define	DOOR_GOLD			3
#define	DOOR_ELEVATOR		4

#define	DOOR_RED			2
#define	DOOR_YELLOW			3
#define	DOOR_GREEN			4
#define	DOOR_BLUE			5
#define	DOOR_GOLDBS			6

#define	DOOR_ONEWAYN		7
#define	DOOR_ONEWAYS		8
#define	DOOR_ONEWAYE		9
#define	DOOR_ONEWAYW		0

#define	SECRET_SIMPLE		1
#define	SECRET_STAY_OPEN	2
#define	SECRET_GL			3

typedef struct
{
	CHAR ID[4];
	LONG cLumps;
	LONG iDir;
} HEADER, *LPHEADER;

typedef struct
{
	LONG iOffset;
	LONG iSize;
	CHAR Name[8];
} DIRECTORY, *LPDIRECTORY;

typedef struct
{
	SHORT x;
	SHORT y;
} VERTEX, *LPVERTEX;

typedef struct
{
	SHORT xOffset;
	SHORT yOffset;
	CHAR szUpper[8];
	CHAR szLower[8];
	CHAR szMiddle[8];
	SHORT Sector;
} SIDEDEF, *LPSIDEDEF;

typedef struct
{
	SHORT Floor;
	SHORT Ceiling;
	CHAR szFloor[8];
	CHAR szCeiling[8];
	SHORT Light;
	SHORT Special;
	SHORT Tag;
} SECTOR, *LPSECTOR;

// Format Independent Classes (03/12/2006)

class CMapThing;
class CMapLinedef;
class CMapLine;

// Original Doom types

typedef struct
{
	SHORT x;
	SHORT y;
	SHORT Angle;
	SHORT Type;
	SHORT Options;
} THING1, *LPTHING1;

typedef struct
{
	SHORT vFrom;
	SHORT vTo;
	SHORT Flags;
	SHORT Types;
	SHORT Tag;
	SHORT sRight;
	SHORT sLeft;
} LINEDEF1, *LPLINEDEF1;

typedef struct
{
	VERTEX vFrom;
	VERTEX vTo;
	SHORT Flags;
	SHORT Types;
	SHORT Tag;
	SIDEDEF sRight;
	SIDEDEF sLeft;
} MAPLINE1, *LPMAPLINE1;

// ZDoom types

typedef struct
{
	SHORT ID;
	SHORT x;
	SHORT y;
	SHORT z;
	SHORT Angle;
	SHORT Type;
	SHORT Options;
	BYTE Special;
	BYTE Args[5];
} THING2, *LPTHING2;

typedef struct
{
	SHORT vFrom;
	SHORT vTo;
	SHORT Flags;
	BYTE Special;
	BYTE Args[5];
	SHORT sRight;
	SHORT sLeft;
} LINEDEF2, *LPLINEDEF2;

typedef struct
{
	VERTEX vFrom;
	VERTEX vTo;
	SHORT Flags;
	BYTE Special;
	BYTE Args[5];
	SIDEDEF sRight;
	SIDEDEF sLeft;
} MAPLINE2, *LPMAPLINE2;

typedef struct tagPOLY_QUEUE
{
	MAPLINE2 Door1;
	MAPLINE2 Door2;
	INT iType;
	INT bDir;
	INT iTexture;
	INT iPoly;
	struct tagPOLY_QUEUE* Next;
} POLY_QUEUE, *LPPOLY_QUEUE;

// Linked lists

typedef struct tagLINE_LIST
{
	CMapLine* lpLine;
	struct tagLINE_LIST* Next;
} LINE_LIST, *LPLINE_LIST;

typedef struct tagSECTOR_LIST
{
	SECTOR Sector;
	struct tagSECTOR_LIST* Next;
} SECTOR_LIST, *LPSECTOR_LIST;

typedef struct tagVERTEX_LIST
{
	VERTEX Vertex;
	struct tagVERTEX_LIST* Next;
} VERTEX_LIST, *LPVERTEX_LIST;

typedef struct tagTHING_LIST
{
	CMapThing* lpThing;
	struct tagTHING_LIST* Next;
} THING_LIST, *LPTHING_LIST;

class CMapConvert
{
protected:
	INT m_iConversionType;

	LPLINE_LIST m_lpList;
	LPSECTOR_LIST m_lpSectors;
	LPVERTEX_LIST m_lpVertices;
	LPTHING_LIST m_lpThings;

	SHORT m_iNextSector;
	USHORT* m_lpAssign;
	USHORT* m_lpSectorTable;

	CHAR m_szCeiling[16];
	CHAR m_szFloor[16];

	// ZDoom specific
	SHORT m_iVertStrip;
	SHORT m_iHorzStrip;
	INT m_xNextDoor;
	INT m_yNextDoor;
	INT m_iNextPoly;
	INT m_iNextSecret;
	LPPOLY_QUEUE m_lpQueue;

	INT m_iLightLevel;

	// Simple or Total Conversion
	BOOL m_bSimpleConversion;

	// Guards/Monsters can open doors (ZDoom)
	BOOL m_bUseAutoDoors;

	// Allows the 16 pixel panel above the textures
	BOOL m_bUseUpperPanels;

	BOOL m_bBuildDoorSockets;

	BOOL m_bAllowFloorZero;
	INT m_iSecretDoorType;

public:
	CMapConvert ();
	~CMapConvert ();

	VOID ResetConversion (VOID);

	INT GetConversionType (VOID);
	VOID SetConversionType (INT iConversionType);

	VOID UseAutoDoors (BOOL bUseAutoDoors);
	VOID AllowFloorZero (BOOL bAllowFloorZero);

	VOID SetLightLevel (INT iLightLevel);

	VOID ChangeFlats (LPSTR lpszCeiling, LPSTR lpszFloor);
	VOID GetFlats (LPSTR lpszCeiling, LPSTR lpszFloor);

	virtual BOOL RunConversion (LPSTR lpszLevel, LPSTR lpszOutput, USHORT* lpFloor, USHORT* lpObject, BOOL bDecorations);

	virtual SHORT GetNewSector (VOID);

	virtual CMapLine* FindLine (LPVERTEX lpvFrom, LPVERTEX lpvTo);
	virtual LPLINE_LIST FindNext (LPVERTEX lpvFrom, LPLINE_LIST* lplpPrev);
	virtual USHORT FindVertex (LPVERTEX lpVertex);
	virtual USHORT FindSector (INT x, INT y);

	virtual VOID SetFrontDoor (INT iType, LPSTR lpszTexture, INT bdir);
	virtual VOID SetBackDoor (INT iType, LPSTR lpszTexture, INT bdir);

	virtual VOID PositionLine (CMapLine* lpLine, INT x, INT y, INT iDir);
	virtual VOID PositionThing (CMapThing* lpThing, INT x, INT y);
	virtual VOID FlipLinedef (CMapLine* lpLine, BOOL bAll = TRUE);
	virtual VOID FillSector (LPSECTOR lpSector);

	virtual VOID AddLine (CMapLine* lpLine);
	virtual VOID AddImpassible (CMapLine* lpLine, USHORT iTexture, BOOL bSecretElevator = FALSE);
	virtual VOID AddSidedLine (CMapLine* lpLine, USHORT iSector, USHORT iTexture);
	virtual VOID AddSecretLine (CMapLine* lpLine, USHORT iSector, USHORT iTexture, INT DoorNumber, BOOL AdjacentToDoor);
	virtual VOID AddSector (LPSECTOR lpSector);
	virtual VOID AddThing (CMapThing* lpThing);

	virtual VOID RemoveLine (CMapLine* lpLine);

protected:
	virtual VOID BuildLineStructure (USHORT* lpFloor, USHORT* lpObject);
	virtual VOID BuildSecretDoors (USHORT* lpFloor, USHORT* lpObject);
	virtual VOID BuildUsableWalls (USHORT* lpFloor, USHORT* lpObject);
	virtual VOID BuildDoor (INT x, INT y, INT iType, INT bDir, INT bSilence);
	virtual VOID BuildDoorBS (INT x, INT y, INT iType, INT bDir, INT bSilence, INT iWallType);
	virtual VOID BuildUsable (INT x, INT y, INT iSwitchTarget, INT iWallType, INT parameter[4]);
	virtual VOID BuildPolyDoor (CMapLine* lpDoor1, CMapLine* lpDoor2, INT iType, INT bDir, INT iTexture, INT iPoly, INT side, INT iWallType);
	virtual VOID BuildHoldingArea (VOID);
	virtual VOID BuildSectors (USHORT* lpFloor);
	virtual VOID BuildWalkover (USHORT* lpFloor, USHORT* lpObject);
	virtual VOID AssignWalkover (INT x, INT y, USHORT* lpObject);
	virtual VOID BuildThings (USHORT* lpObject, USHORT* lpFloor);
	virtual VOID BuildPlayerStarts (USHORT* lpObject, USHORT* lpFloor);
	virtual VOID BuildDMPlayerStarts (USHORT* lpObject, USHORT* lpFloor);
	virtual INT CheckThingPos (INT x, INT y, INT dx, INT dy, USHORT* lpObject, USHORT* lpFloor);
	virtual VOID ConvertThing_Simple (CMapThing* lpThing, INT x, INT y, USHORT iType, BOOL bDeaf);
	virtual VOID ConvertThing_TotalConversion (USHORT* lpObject, CMapThing* lpThing, INT x, INT y, USHORT iType, BOOL bDeaf, INT iFloor);
	virtual INT CheckThingNeighbor (USHORT* lpObject, INT iType, INT x, INT y, INT from);
	virtual BOOL TranslateActor (CMapThing* lpThing, USHORT iWolfType);
	virtual VOID ConvertFinale (USHORT* lpObject);
	virtual BOOL AssignSector (INT x, INT y);
	virtual VOID MergeDone (CMapLine* lpLine);
	virtual VOID MergeLines (VOID);
	virtual VOID BuildOutsideViews (VOID);
	virtual VOID BuildPrefabs (USHORT* lpObject);
	virtual VOID BuildPolys (VOID);
	virtual VOID FixSidedefs (VOID);
	virtual VOID BuildVertexList (VOID);
	virtual VOID BuildWadFile (LPSTR lpszMapName, LPSTR lpszFile, BOOL bDecorations);

	virtual ULONG SerializeThings (HANDLE hFile);
	virtual ULONG SerializeLinedefs (HANDLE hFile);
	virtual ULONG SerializeSidedefs (HANDLE hFile);
	virtual ULONG SerializeVertices (HANDLE hFile);
	virtual ULONG SerializeSectors (HANDLE hFile);
	virtual ULONG SerializeLump (HANDLE hFile, LPSTR lpszFile);
};

#endif