#include <windows.h>
#include "resource.h"
#include "..\MapLine.h"
#include "SelectFlatsDlg.h"
#include "OpenWolfDlg.h"
#include "OptionsDlg.h"
#include "LightLevelDlg.h"
#include "OtherOptionsDlg.h"
#include "WinConvApp.h"

CWinConvApp::CWinConvApp (HINSTANCE hInstance, LPSTR lpszCmdLine)
{
	m_lpWolfMap = new CWolfMap;

	m_hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
	m_hThread = NULL;

	lstrcpy(m_szCeiling,"FLAT19");
	lstrcpy(m_szFloor,"FLOOR7_1");

	if(!GetWolfFilesFromRegistry(m_szHead,m_szData))
	{
		m_szHead[0] = 0;
		m_szData[0] = 0;
	}

	m_xCenter = 0;
	m_yCenter = 0;
	m_fScale = 0.08f;

	m_hbmBuffer = NULL;
	m_hdcBuffer = NULL;
	m_hbrWhite = (HBRUSH)GetStockObject(WHITE_BRUSH);
	m_hpnBlack = CreatePen(PS_SOLID,0,RGB(0,0,0));
	m_hpnBlue = CreatePen(PS_SOLID,2,RGB(50,50,255));

	m_iPhase = 0;
	m_lpPhase1 = NULL; m_lpEnd1 = NULL;
	m_lpPhase2 = NULL; m_lpEnd2 = NULL;

	InitApp(hInstance,"Wolf Map To Wad",lpszCmdLine,"MapToWadCls");
}

CWinConvApp::~CWinConvApp ()
{
	DeleteObject(m_hpnBlue);

	if(m_hThread)
		CloseHandle(m_hThread);
	CloseHandle(m_hEvent);

	ClearDrawLists();

	DeleteObject(m_hpnBlue);
	DeleteObject(m_hpnBlack);
	DeleteDC(m_hdcBuffer);
	DeleteObject(m_hbmBuffer);

	delete m_lpWolfMap;
}

BOOL CWinConvApp::IsConverting (VOID)
{
	if(WaitForSingleObject(m_hEvent,0) == WAIT_TIMEOUT)
		return FALSE;
	return TRUE;
}

//

BOOL CWinConvApp::onCreateWindow (HWND hwnd)
{
	return TRUE;
}

BOOL CWinConvApp::onFinalize (VOID)
{
	return TRUE;
}

VOID CWinConvApp::onDestroy (VOID)
{
}

BOOL CWinConvApp::onClose (BOOL bTopLevel)
{
	onCommand(IDM_EXIT,0);
	return TRUE;
}

VOID CWinConvApp::onActivate (BOOL bActive)
{
}

VOID CWinConvApp::onGainFocus (HWND hwndLosing)
{
}

VOID CWinConvApp::onSizeWindow (INT iWidth, INT iHeight)
{
	HDC hdc;

	m_xOffset = iWidth / 2;
	m_yOffset = iHeight / 2;

	DeleteDC(m_hdcBuffer);
	DeleteObject(m_hbmBuffer);
	hdc = GetDC();
	m_hbmBuffer = CreateCompatibleBitmap(hdc,iWidth,iHeight);
	m_hdcBuffer = CreateCompatibleDC(hdc);
	SelectObject(m_hdcBuffer,m_hbmBuffer);
	ReleaseDC(hdc);
}

VOID CWinConvApp::onMouseMove (INT iKeys, INT x, INT y)
{
}

VOID CWinConvApp::onLButtonDown (INT iKeys, INT x, INT y)
{
}

VOID CWinConvApp::onLButtonUp (INT x, INT y)
{
}

VOID CWinConvApp::onLButtonDbl (INT x, INT y)
{
}

VOID CWinConvApp::onRButtonDown (INT iKeys, INT x, INT y)
{
}

VOID CWinConvApp::onRButtonUp (INT x, INT y)
{
}

VOID CWinConvApp::onRButtonDbl (INT x, INT y)
{
}

VOID CWinConvApp::onKeyDown (INT iKey)
{
	BOOL bRedraw = FALSE;
	switch(iKey)
	{
	case VK_LEFT:
		ShiftCenter(8,0);
		bRedraw = TRUE;
		break;
	case VK_RIGHT:
		ShiftCenter(-8,0);
		bRedraw = TRUE;
		break;
	case VK_UP:
		ShiftCenter(0,8);
		bRedraw = TRUE;
		break;
	case VK_DOWN:
		ShiftCenter(0,-8);
		bRedraw = TRUE;
		break;
	}
	if(bRedraw)
		InvalidateRect(m_hwnd,NULL,TRUE);
}

VOID CWinConvApp::onKeyUp (INT iKey)
{
}

VOID CWinConvApp::onChar (INT ch)
{
	BOOL bRedraw = FALSE;
	if(ch == '+' || ch == '=')
	{
		SetScale(GetScale() * (FLOAT)1.2);
		bRedraw = TRUE;
	}
	else if(ch == '-')
	{
		SetScale(GetScale() / (FLOAT)1.2);
		bRedraw = TRUE;
	}
	if(bRedraw)
		InvalidateRect(m_hwnd,NULL,TRUE);
}

LRESULT CWinConvApp::onCommand (WPARAM wParam, LPARAM lParam)
{
	LRESULT lResult = 0;
	switch(LOWORD(wParam))
	{
	case IDM_SELECT_FLATS:
		if(IsConverting() == FALSE)
		{
			CSelectFlatsDlg* lpSelectFlatsDlg = new CSelectFlatsDlg(m_hInstance);
			lpSelectFlatsDlg->SetFlats(m_szFloor,m_szCeiling);
			if(lpSelectFlatsDlg->Display(m_hwnd))
				lpSelectFlatsDlg->GetFlats(m_szFloor,m_szCeiling);
			delete lpSelectFlatsDlg;
		}
		break;
	case IDM_LIGHT_LEVEL:
		if(IsConverting() == FALSE)
		{
			CLightLevelDlg* lpLightLevelDlg = new CLightLevelDlg(m_hInstance);
			lpLightLevelDlg->SetLightLevel(m_iLightLevel);
			if(lpLightLevelDlg->Display(m_hwnd))
				m_iLightLevel = lpLightLevelDlg->GetLightLevel();
			delete lpLightLevelDlg;
		}
		break;
	case IDM_OTHER_OPTIONS:
		if(IsConverting() == FALSE)
		{
			COtherOptionsDlg* lpOtherOptionsDlg = new COtherOptionsDlg(m_hInstance);
			lpOtherOptionsDlg->SetUpperPanels(m_bUseUpperPanels);
			lpOtherOptionsDlg->SetDoorSockets(m_bBuildDoorSockets);
			if(lpOtherOptionsDlg->Display(m_hwnd))
			{
				m_bUseUpperPanels = lpOtherOptionsDlg->GetUpperPanels();
				m_bBuildDoorSockets = lpOtherOptionsDlg->GetDoorSockets();
			}
			delete lpOtherOptionsDlg;
		}
		break;
	case IDM_OPEN_WOLF:
		if(IsConverting() == FALSE)
		{
			COpenWolfDlg* lpOpenWolfDlg = new COpenWolfDlg(m_hInstance,m_lpWolfMap);
			lpOpenWolfDlg->SetPaths(m_szHead,m_szData);
			if(lpOpenWolfDlg->Display(m_hwnd))
			{
				lpOpenWolfDlg->GetPaths(m_szHead,m_szData);
				if(m_lpWolfMap->Init(m_szHead,m_szData))
					SetWolfFilesToRegistry(m_szHead,m_szData);
				else
					MessageBox(m_hwnd,"There were problems opening the files you selected.","File Error",MB_OK);
			}
			delete lpOpenWolfDlg;
		}
		break;
	case IDM_CONVERT:
		if(IsConverting() == FALSE)
		{
			if(m_lpWolfMap->IsOpen())
			{
				COptionsDlg* lpOptionsDlg = new COptionsDlg(m_hInstance,m_lpWolfMap);
				lpOptionsDlg->SetConversionType(m_iConversionType);
				lpOptionsDlg->SetFullConversion(!m_bSimpleConversion);
				if(lpOptionsDlg->Display(m_hwnd))
				{
					DWORD dwThread;
					m_iConversionType = lpOptionsDlg->GetConversionType();
					lpOptionsDlg->GetMap(m_szMap);
					lpOptionsDlg->GetLevel(m_szLevel);
					lpOptionsDlg->GetOutput(m_szOutput);

					m_bUseAutoDoors = lpOptionsDlg->UseAutoDoors();
					m_bAllowFloorZero = lpOptionsDlg->AllowFloorZero();
					m_bSimpleConversion = !lpOptionsDlg->GetFullConversion();

					if(m_lpWolfMap->GetMap(m_szMap))
					{
						ClearDrawLists();
						InvalidateRect(m_hwnd,NULL,TRUE);
						SetEvent(m_hEvent);
						m_hThread = CreateThread(NULL,0,ConversionThread,this,0,&dwThread);
					}
					else
					{
						CHAR szError[256];
						wsprintf(szError,"Could not read map \"%s\" from data file",m_szMap);
						MessageBox(m_hwnd,szError,"Map Error",MB_OK);
					}
				}
				delete lpOptionsDlg;
			}
			else
				MessageBox(m_hwnd,"You must open a Wolf map first","MapToWad",MB_OK);
		}
		break;
	case IDM_EXIT:
		if(IsConverting() == FALSE)
			PostQuitMessage(0);
		else
			MessageBox(m_hwnd,"You must wait for the conversion to complete","Conversion in Progress",MB_OK);
		break;
	}
	return lResult;
}

VOID CWinConvApp::onPaint (PAINTSTRUCT* lpps, HDC hdc)
{
	HPEN hDefault = (HPEN)SelectObject(m_hdcBuffer,m_hpnBlack);

	FillRect(m_hdcBuffer,&m_rcWindow,m_hbrWhite);

	DrawList(m_hdcBuffer,m_lpPhase1);
	SelectObject(m_hdcBuffer,m_hpnBlue);
	DrawList(m_hdcBuffer,m_lpPhase2);
	SelectObject(m_hdcBuffer,hDefault);

	BitBlt(hdc,0,0,m_nWidth,m_nHeight,m_hdcBuffer,0,0,SRCCOPY);
}

BOOL CWinConvApp::onPaintBackground (HDC hdc)
{
	return FALSE;
}

VOID CWinConvApp::onDraw (VOID)
{
}

VOID CWinConvApp::onProcess (VOID)
{
}

VOID CWinConvApp::onMakeWndClass (WNDCLASS& wc)
{
	CQuickWin::onMakeWndClass(wc);
}

BOOL CWinConvApp::onBeforeCreateWin (DWORD& dwExStyle, DWORD& dwStyle)
{
	return CQuickWin::onBeforeCreateWin(dwExStyle,dwStyle);
}

BOOL CWinConvApp::onWinMessage (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam, LRESULT& lResult)
{
	if(message == WM_ADD_DRAW_ITEM)
	{
		LPDRAWLIST lpItem = (LPDRAWLIST)lParam;
		HDC hdc = GetDC();
		HGDIOBJ hDefault = NULL;
		LPDRAWLIST* lplpList, *lplpEnd;
		if(wParam == 1)
		{
			hDefault = SelectObject(hdc,m_hpnBlack);
			lplpList = &m_lpPhase1;
			lplpEnd = &m_lpEnd1;
		}
		else if(wParam == 2)
		{
			hDefault = SelectObject(hdc,m_hpnBlue);
			lplpList = &m_lpPhase2;
			lplpEnd = &m_lpEnd2;
		}
		if(*lplpEnd)
			(*lplpEnd)->Next = lpItem;
		else
			(*lplpList) = lpItem;
		(*lplpEnd) = lpItem;
		DrawList(hdc,lpItem);
		SelectObject(hdc,hDefault);
		ReleaseDC(hdc);
		return TRUE;
	}
	else if(message == WM_END_CONVERSION)
	{
		CloseHandle(m_hThread);
		m_hThread = NULL;
		ResetEvent(m_hEvent);
		MessageBox(m_hwnd,"Conversion is Complete!","Done",MB_OK);
		return TRUE;
	}
	return FALSE;
}

//

BOOL CWinConvApp::onStartApp (HINSTANCE hInstance, LPSTR lpszCmdLine)
{
	return TRUE;
}

HMENU CWinConvApp::onLoadMenu (HINSTANCE hInstance)
{
	return LoadMenu(hInstance,MAKEINTRESOURCE(IDR_MENU));
}

BOOL CWinConvApp::onInitApp (HWND hwnd, BOOL bFullScreen)
{
	return TRUE;
}

VOID CWinConvApp::onProcessApp (VOID)
{
}

VOID CWinConvApp::onExitProgram (VOID)
{
}

//

VOID CWinConvApp::AddLine (CMapLine* lpLine)
{
	CMapConvert::AddLine(lpLine);
	AddLineItem(lpLine,1);
}

BOOL CWinConvApp::TranslateActor (CMapThing* lpThing, USHORT iWolfType)
{
	return CMapConvert::TranslateActor(lpThing,iWolfType);
}

BOOL CWinConvApp::AssignSector (INT x, INT y)
{
	if(CMapConvert::AssignSector(x,y))
	{
		LPDRAWLIST lpItem = new DRAWLIST;
		lpItem->iType = 2;
		lpItem->x1 = x;
		lpItem->y1 = y;
		lpItem->x2 = -1;
		lpItem->y2 = -1;
		lpItem->Next = NULL;
		PostMessage(WM_ADD_DRAW_ITEM,1,(LPARAM)lpItem);
		return TRUE;
	}
	return FALSE;
}

VOID CWinConvApp::MergeDone (CMapLine* lpLine)
{
	AddLineItem(lpLine,2);
}

VOID CWinConvApp::ShiftCenter (INT xShift, INT yShift)
{
	SetCenter(m_xCenter + (INT)((FLOAT)xShift / m_fScale),m_yCenter + (INT)((FLOAT)yShift / m_fScale));
}

VOID CWinConvApp::SetCenter (INT x, INT y)
{
	m_xCenter = x;
	m_yCenter = y;
}

VOID CWinConvApp::GetCenter (INT* lpx, INT* lpy)
{
	if(lpx)
		*lpx = m_xCenter;
	if(lpy)
		*lpy = m_yCenter;
}

VOID CWinConvApp::SetScale (FLOAT fScale)
{
	if(fScale >= (FLOAT).001)
	{
		if(fScale <= 4)
			m_fScale = fScale;
		else
			m_fScale = 4;
	}
	else
		fScale = (FLOAT).001;
}

FLOAT CWinConvApp::GetScale (VOID)
{
	return m_fScale;
}

VOID CWinConvApp::AddLineItem (CMapLine* lpLine, INT iPhase)
{
	LPDRAWLIST lpItem;
	lpItem = new DRAWLIST;
	lpItem->iType = 1;
	lpItem->x1 = lpLine->m_vFrom.x;
	lpItem->y1 = lpLine->m_vFrom.y;
	lpItem->x2 = lpLine->m_vTo.x;
	lpItem->y2 = lpLine->m_vTo.y;
	lpItem->Next = NULL;
	PostMessage(WM_ADD_DRAW_ITEM,iPhase,(LPARAM)lpItem);
}

VOID CWinConvApp::DrawList (HDC hdc, LPDRAWLIST lpList)
{
	LONG x, y;
	while(lpList)
	{
		if(lpList->iType == 1)
		{
			x = (INT)((FLOAT)(lpList->x1 + m_xCenter) * m_fScale);
			y = (INT)((FLOAT)(-1 * lpList->y1 + m_yCenter) * m_fScale);
			MoveToEx(hdc,x + m_xOffset,y + m_yOffset,NULL);

			x = (INT)((FLOAT)(lpList->x2 + m_xCenter) * m_fScale);
			y = (INT)((FLOAT)(-1 * lpList->y2 + m_yCenter) * m_fScale);
			LineTo(hdc,x + m_xOffset,y + m_yOffset);

			SetPixelV(hdc,x + m_xOffset,y + m_yOffset,0);
		}
		else if(lpList->iType == 2)
		{
			x = lpList->x1 * 64 - 2048;
			y = 2048 - (lpList->y1 + 1) * 64;

			x = (INT)((FLOAT)(x + m_xCenter) * m_fScale);
			y = (INT)((FLOAT)(-1 * y + m_yCenter) * m_fScale);
			MoveToEx(hdc,x + m_xOffset,y + m_yOffset,NULL);

			x = (lpList->x1 + 1) * 64 - 2048;
			y = 2048 - lpList->y1 * 64;

			x = (INT)((FLOAT)(x + m_xCenter) * m_fScale);
			y = (INT)((FLOAT)(-1 * y + m_yCenter) * m_fScale);
			LineTo(hdc,x + m_xOffset,y + m_yOffset);
		}
		lpList = lpList->Next;
	}
}

VOID CWinConvApp::ClearDrawLists (VOID)
{
	LPDRAWLIST lpTemp;
	while(m_lpPhase1)
	{
		lpTemp = m_lpPhase1;
		m_lpPhase1 = m_lpPhase1->Next;
		delete lpTemp;
	}
	m_lpEnd1 = NULL;
	while(m_lpPhase2)
	{
		lpTemp = m_lpPhase2;
		m_lpPhase2 = m_lpPhase2->Next;
		delete lpTemp;
	}
	m_lpEnd2 = NULL;
	m_iPhase = 0;
}

BOOL CWinConvApp::GetWolfFilesFromRegistry (LPSTR lpszHead, LPSTR lpszData)
{
	BOOL bSuccess = FALSE;
	if(lpszHead && lpszData)
	{
		HKEY hKey = NULL;
		if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Simbey\\MapToWad",0,KEY_READ,&hKey) == ERROR_SUCCESS)
		{
			DWORD dwSize;

			*lpszHead = 0;
			*lpszData = 0;

			dwSize = 256;
			RegQueryValueEx(hKey,"MapHead",NULL,NULL,(LPBYTE)lpszHead,&dwSize);

			dwSize = 256;
			RegQueryValueEx(hKey,"GameMaps",NULL,NULL,(LPBYTE)lpszData,&dwSize);

			RegCloseKey(hKey);

			bSuccess = TRUE;
		}
	}
	return bSuccess;
}

BOOL CWinConvApp::SetWolfFilesToRegistry (LPSTR lpszHead, LPSTR lpszData)
{
	BOOL bSuccess = FALSE;
	if(lpszHead && lpszData)
	{
		HKEY hKey = NULL;
		if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Simbey\\MapToWad",0,KEY_WRITE,&hKey) != ERROR_SUCCESS)
		{
			DWORD dwDisposition = 0;
			RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Simbey\\MapToWad",0,NULL,REG_OPTION_NON_VOLATILE,KEY_WRITE,NULL,&hKey,&dwDisposition);
		}
		if(hKey)
		{
			RegSetValueEx(hKey,"MapHead",0,REG_SZ,(LPBYTE)lpszHead,lstrlen(lpszHead) + 1);
			RegSetValueEx(hKey,"GameMaps",0,REG_SZ,(LPBYTE)lpszData,lstrlen(lpszData) + 1);

			RegCloseKey(hKey);

			bSuccess = TRUE;
		}
	}
	return bSuccess;
}

DWORD WINAPI CWinConvApp::ConversionThread (LPVOID lpParam)
{
	CWinConvApp* lpApp = (CWinConvApp*)lpParam;
	LPMAPDATA lpMap = lpApp->m_lpWolfMap->GetMap(lpApp->m_szMap);
	if(lpMap)
	{
		lpApp->m_iPhase = 1;
		lpApp->ResetConversion();
		lpApp->RunConversion(lpApp->m_szLevel,lpApp->m_szOutput,lpMap->Floor,lpMap->Object,TRUE);
		lpApp->m_iPhase = 0;
		lpApp->PostMessage(WM_END_CONVERSION,0,0);
	}
	ExitThread(0);
	return 0;
}