// 10/19/2003 - Finished Wolf/Spear map reading
// 10/21/2004 - Added map name enumeration
// 03/28/2006 - Added ability to read MAPHEAD files from FloEdit
// 08/03/2008 - Added support for ChaosEdit MAP files
// 08/05/2008 - Added support for SD2 and SD3 files

#include <windows.h>
#include "Expand.h"
#include "WolfMap.h"

CWolfMap::CWolfMap ()
{
	ZeroMemory(&m_Head,sizeof(MAPHEAD));
	m_cData = 0;
	m_lpData = NULL;
	m_bRawMap = FALSE;
}

CWolfMap::~CWolfMap ()
{
	delete m_lpData;
}

BOOL CWolfMap::IsOpen (VOID)
{
	return (m_lpData) ? TRUE : FALSE;
}

BOOL CWolfMap::Init (LPSTR lpszHead, LPSTR lpszData)
{
	BOOL bSuccess = FALSE;
	if(m_lpData)
	{
		delete m_lpData;
		m_lpData = NULL;
		m_bRawMap = FALSE;
	}
	if(LoadMapHead(lpszHead,&m_Head))
	{
		HANDLE hFile = CreateFile(lpszData,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
		if(hFile != INVALID_HANDLE_VALUE)
		{
			if(m_bRawMap)
			{
				Format::Type eType = IsRawMap(lpszData,hFile);
				if(eType == Format::Raw || eType == Format::RawWithHeader)
				{
					m_cData = sizeof(MAPTYPE) + 8 + 16384;
					m_lpData = new BYTE[m_cData];
					if(m_lpData)
					{
						DWORD dwRead = 0;
						LPMAPTYPE lpType = (LPMAPTYPE)(m_lpData + 8);
						lpType->Width = 64;
						lpType->Height = 64;
						if(eType == Format::RawWithHeader)
						{
							CHAR szHeader[34];
							ReadFile(hFile,szHeader,34,&dwRead,NULL);
							CopyMemory(lpType->Name,szHeader + 18,16);
						}
						else
							lstrcpy(lpType->Name,"Custom Map");
						lpType->PlaneLength[0] = 8192;
						lpType->PlaneLength[1] = 8192;
						lpType->PlaneLength[2] = 0;
						lpType->PlaneStart[0] = 8 + sizeof(MAPTYPE);
						lpType->PlaneStart[1] = 8 + sizeof(MAPTYPE) + 8192;
						ReadFile(hFile,m_lpData + 8 + sizeof(MAPTYPE),16384,&dwRead,NULL);
						if(dwRead == 16384)
							bSuccess = TRUE;
					}
				}
				else
					m_bRawMap = FALSE;
			}
			else
			{
				m_cData = GetFileSize(hFile,NULL);
				m_lpData = new BYTE[m_cData];
				if(m_lpData)
				{
					DWORD dwRead = 0;
					ReadFile(hFile,m_lpData,m_cData,&dwRead,NULL);
					if(dwRead == m_cData)
						bSuccess = TRUE;
				}
			}
		}
	}
	return bSuccess;
}

USHORT CWolfMap::TotalMaps (VOID)
{
	USHORT c = 0, i;
	if(m_lpData)
	{
		for(i = 0; i < 100; i++)
		{
			if(m_Head.MapStart[i] != 0)
				c++;
		}
	}
	return c;
}

LONG CWolfMap::EnumMaps (ENUMWOLFMAPS lpfnEnum, LPVOID lpParam)
{
	LONG cEnum = 0;
	if(m_lpData && lpfnEnum)
	{
		LPMAPTYPE lpType;
		LONG i, iOffset;
		if(m_bRawMap)
		{
			iOffset = m_Head.MapStart[0];
			lpType = (LPMAPTYPE)(m_lpData + iOffset);
			lpfnEnum(lpType,lpParam);
			cEnum = 1;
		}
		else
		{
			for(i = 0; i < 100; i++)
			{
				iOffset = m_Head.MapStart[i];
				if(iOffset == 0)
					continue;
				if(iOffset + sizeof(MAPTYPE) <= m_cData)
				{
					cEnum++;
					lpType = (LPMAPTYPE)(m_lpData + iOffset);
					if(lpfnEnum(lpType,lpParam) == FALSE)
						break;
				}
			}
		}
	}
	return cEnum;
}

LPMAPDATA CWolfMap::GetMap (LPSTR lpszMap)
{
	LPMAPDATA lpMap = NULL;
	if(m_lpData)
	{
		if(lpszMap)
		{
			LPMAPTYPE lpType;
			LONG i, iOffset;
			for(i = 0; i < 100; i++)
			{
				iOffset = m_Head.MapStart[i];
				if(iOffset == 0)
					continue;
				if(iOffset + sizeof(MAPTYPE) <= m_cData)
				{
					lpType = (LPMAPTYPE)(m_lpData + iOffset);
					if(CompareMapName(lpType,lpszMap))
					{
						lpMap = ExpandMap(&m_Head,lpType,m_lpData);
						break;
					}
				}
			}
		}
	}
	return lpMap;
}

LPMAPDATA CWolfMap::GetMap (USHORT iMapNum)
{
	LPMAPDATA lpMap = NULL;
	if(m_lpData)
	{
		if(iMapNum < 100)
		{
			LPMAPTYPE lpType;
			LONG iOffset = m_Head.MapStart[iMapNum];
			if(iOffset > 0)
			{
				if(iOffset + sizeof(MAPTYPE) <= m_cData)
				{
					lpType = (LPMAPTYPE)(m_lpData + iOffset);
					lpMap = ExpandMap(&m_Head,lpType,m_lpData);
				}
			}
		}
	}
	return lpMap;
}

VOID CWolfMap::FreeMap (LPMAPDATA lpMap)
{
	delete lpMap;
}

//

BOOL CWolfMap::LoadMapHead (LPSTR lpszFile, LPMAPHEAD lpMapHead)
{
	BOOL bSuccess = FALSE;
	HANDLE hFile = CreateFile(lpszFile,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(hFile != INVALID_HANDLE_VALUE)
	{
		Format::Type eType = IsRawMap(lpszFile,hFile);
		if(eType == Format::Raw || eType == Format::RawWithHeader)
		{
			ZeroMemory(lpMapHead,sizeof(MAPHEAD));
			lpMapHead->MapStart[0] = 8;	// It has to be greater than zero
			lpMapHead->sEncoding = 0;
			m_bRawMap = TRUE;
			bSuccess = TRUE;
		}
		else if(eType == Format::Normal)
		{
			DWORD dwTemp = 0;
			ReadFile(hFile,&lpMapHead->sEncoding,2,&dwTemp,NULL);
			ReadFile(hFile,lpMapHead->MapStart,400,&dwTemp,NULL);
			if(dwTemp == 400)
				bSuccess = TRUE;
			else if(dwTemp == 240)	// This is actually 60 levels
			{
				ZeroMemory((LPBYTE)lpMapHead->MapStart + 240,160);
				bSuccess = TRUE;
			}
			else if(dwTemp == 84)	// ???
			{
				ZeroMemory((LPBYTE)lpMapHead->MapStart + 84,316);
				bSuccess = TRUE;
			}
		}
		CloseHandle(hFile);
	}
	return bSuccess;
}

Format::Type CWolfMap::IsRawMap (LPSTR lpszFile, HANDLE hFile)
{
	Format::Type eType = Format::Normal;
	if(lpszFile && hFile != INVALID_HANDLE_VALUE)
	{
		LPSTR lpszExt = strrchr(lpszFile,'.');
		if(lpszExt && lstrcmpi(lpszExt,".MAP") == 0)
		{
			DWORD dwFileSize = GetFileSize(hFile,NULL);
			if(dwFileSize == 16384)
				eType = Format::Raw;
			else if(dwFileSize == 16418)
				eType = Format::RawWithHeader;
			else
				eType = Format::Bad;
		}
	}
	return eType;
}

LPMAPDATA CWolfMap::ExpandMap (LPMAPHEAD lpMapHead, LPMAPTYPE lpType, LPBYTE lpFile)
{
	LPMAPDATA lpMap = new MAPDATA;
	if(lpMap)
	{
		USHORT* lpSrc, *lpDest;
		INT i;
		try
		{
			if(m_bRawMap)
			{
				CopyMemory(lpMap->Floor,lpFile + lpType->PlaneStart[0],8192);
				CopyMemory(lpMap->Object,lpFile + lpType->PlaneStart[1],8192);
				RotateData(lpMap->Floor);
				RotateData(lpMap->Object);
			}
			else
			{
				for(i = 0; i < 2; i++)
				{
					lpSrc = (USHORT*)(lpFile + lpType->PlaneStart[i]);
					if(i == 0)
						lpDest = lpMap->Floor;
					else if(i == 1)
						lpDest = lpMap->Object;
					if(*lpSrc == 8192)
					{
						lpSrc++;	// skip length
						CExpand::RLEWexpand(lpSrc,lpDest,8192,lpMapHead->sEncoding);
					}
					else
					{
						USHORT Expand[4096];
						USHORT Size = *lpSrc;
						CExpand::CarmackExpand(lpSrc+1,Expand,Size);
						CExpand::RLEWexpand(Expand+1,lpDest,8192,lpMapHead->sEncoding);
					}
				}
			}
		}
		catch(...)
		{
			delete lpMap;
			lpMap = NULL;
		}
	}
	return lpMap;
}

VOID CWolfMap::RotateData (USHORT* lpData)
{
	INT x, y;
	USHORT Rotated[4096];
	for(y = 63; y >= 0; y--)
	{
		for(x = 0; x < 64; x++)
			Rotated[x * 64 + y] = lpData[y * 64 + x];
	}
	CopyMemory(lpData,Rotated,8192);
}

BOOL CWolfMap::CompareMapName (LPMAPTYPE lpDesc, LPSTR lpszName)
{
	BOOL bMatch = TRUE;
	LPSTR lpszField = lpDesc->Name;
	INT i;
	CHAR ch;
	for(i = 0; i < 16; i++)
	{
		ch = toupper(lpszField[i]);
		if(toupper(lpszName[i]) != ch)
		{
			bMatch = FALSE;
			break;
		}
		if(ch == 0)
			break;
	}
	return bMatch;
}